//
// Ucs.cpp
//
// Otfried Cheong
//
// Encode/decode for UCS2
// 

#include "ufcodec.h"

#include <e32def.h>
#include <e32uid.h>

// ------------------------------------------------------------

class Uni {
public:
  enum { KIllegal = 0xffff, KIncomplete = 0xf800, KHeadless = 0xf801 };
  static TText16 WChar(const TDesC8 &aDes, TInt &aPos);
  static void AppendUTF8(TDes8 &aUtf, TText16 aWChar);
};

LOCAL_D const TUint8 bytesFromUTF8[256] = {
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1, 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
  2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2, 3,3,3,3,3,3,3,3,4,4,4,4,5,5,5,5};

LOCAL_D const TText16 firstByteMark[7] = {
  0x00, 0xC0, 0xE0, 0xF0, 0xF8, 0xFC, 0 };

// convert UTF8 to Unicode
// special cases:  end of string in UTF8 sequence --> KIncomplete
//                 UTF8 sequence stops prematurely --> KIllegal
//                 UTF8 sequence starts with trailer --> KHeadless
TText16 Uni::WChar(const TDesC8 &aDes, TInt &aPos)
{
  TUint wch = aDes[aPos++];
  if ((wch & 0xc0) == 0x80) {
    // KHeadless
    while (aPos < aDes.Length() && (aDes[aPos] & 0xc0) == 0x80)
      aPos++;
    return KHeadless;
  }
  TInt extraBytes = bytesFromUTF8[wch & 0xff];
  wch -= firstByteMark[extraBytes];
  while (extraBytes--) {
    if (aPos >= aDes.Length())
      return TText16(KIncomplete);
    if ((aDes[aPos] & 0xc0) != 0x80)
      // return TText16(aDes[aPos - 1]);
      return TText16(KIllegal);
    wch <<= 6;
    wch |= aDes[aPos++] & 0x3f; 
  }
  return TText16(wch);
}

// convert Unicode to UTF8
void Uni::AppendUTF8(TDes8 &aUtf, TText16 wch)
{
  if (wch < 0x80) {
    aUtf.Append(TChar(wch));
  } else if (wch < 0x800) {
    aUtf.Append(TChar(((wch & 0x7c0) >> 6) | 0xc0));
    aUtf.Append(TChar((wch & 0x03f) | 0x80));
  } else {
    // We never need to write UCS larger than 0x10000,
    // so now it's three bytes 
    aUtf.Append(TChar(((wch & 0x0f000) >> 12) | 0xe0));
    aUtf.Append(TChar(((wch & 0xfc0) >> 6) | 0x80));
    aUtf.Append(TChar((wch & 0x03f) | 0x80));
  }
}

// ------------------------------------------------------------

class CUcsEncoder : public CUniCodec {
public:
  CUcsEncoder() { iChar = -1; }

  virtual TText16 DecodeWChar(TText16 aChar) { return aChar; }
  virtual TText16 EncodeWChar(TText16 aChar) { return aChar; }
  virtual HBufC8 *EncodeTextL(const TDesC8 &aText, TBool aSkipUnknown);
  virtual HBufC8 *DecodeTextL(const TDesC8 &aText);

  virtual TUid Uid() const { return TUid::Uid(0x10000e27); }
  virtual TInt ApiLevel() const { return 17; }
  virtual TTime Version() const
  { return TTime(TDateTime(2000, ENovember, 1, 0, 0, 0, 0)); }
  virtual TInt MaximumSizeInBytes() const { return 2; }
  virtual TInt MaximumSizeOfUnicodeRepresentation() const { return 1; }
  virtual void EncodingContextReset() { }
  virtual TBool EncodingContextNonTrivial() const { return EFalse; }
  virtual TInt EncodingContextData() const { return 0; }
  virtual void DecodingContextReset() { iChar = -1; } 
  virtual TBool DecodingContextNonTrivial() const { return (iChar >= 0); }
  virtual TInt DecodingContextData() const 
  { return (iChar >= 0) ? 1 : 0; }

  virtual void EncodeL(MUniCodecUnicodeSource *source,
		       MUniCodecNativeSink &sink);
  virtual TInt DecodeL(MUniCodecNativeSource *source,
                       MUniCodecUnicodeSink &sink);
private:
  TInt iChar;
};

// --------------------------------------------------------------------

HBufC8* CUcsEncoder::EncodeTextL(const TDesC8 &aText, TBool)
{
  HBufC8 *anEncoded = HBufC8::NewL(2 * aText.Length());
  TPtr8 aBuf = anEncoded->Des();
  TInt i = 0;
  while (i < aText.Length()) {
    TText16 aUni = Uni::WChar(aText, i);
    aBuf.Append(TChar(aUni >> 8));
    aBuf.Append(TChar(aUni & 0xff));
  }
  return anEncoded;
}

HBufC8* CUcsEncoder::DecodeTextL(const TDesC8 &aText)
{
  HBufC8 *aDecoded = HBufC8::NewL((3 * aText.Length() + 2)/2);
  TPtr8 aBuf = aDecoded->Des();
  TInt j = 0;
  TUint aUni;
  while (j < aText.Length() - 1) {
    aUni = aText[j++] << 8;
    aUni |= aText[j++];
    Uni::AppendUTF8(aBuf, aUni);
  }
  return aDecoded;
}

// ------------------------------------------------------------

void CUcsEncoder::EncodeL(MUniCodecUnicodeSource *source, 
			  MUniCodecNativeSink &sink)
{
  if (!source)
    // no context to flush
    return;

  for (;;) {
    TInt wch = source->GetL();
    if (wch < 0) 
      return;
    sink.PutL(wch >> 8);
    if (sink.PutL(wch & 0xff))
      return;
  }
}

TInt CUcsEncoder::DecodeL(MUniCodecNativeSource *source, 
			  MUniCodecUnicodeSink &sink)
{
  if (!source) {
    if (iChar >= 0)
      sink.IllegalL(iChar);
    return 0;
  }
  
  for (;;) {
    if (iChar < 0) {
      TInt ch = source->GetL(ETrue);
      if (ch < 0)
	return 0;
      iChar = ch << 8;
    } else {
      // iChar >= 0
      TInt ch = source->GetL(EFalse);
      if (ch < 0)
	return 1;
      ch |= iChar;
      iChar = -1;
      if (sink.PutL(ch))
	return 0;
    }
  }
}

// ---------------------------------------------------------------------

// The E32Dll() entry point function

GLDEF_C TInt E32Dll(TDllReason /*aReason*/)
{
  return(KErrNone);
}

// Function to construct a CUniCodec object. Note that this function
// is exported at ordinal 1.

EXPORT_C CUniEncoder* NewUcsEncoder()
{
  return new CUcsEncoder;
}
