//
// UfcKsc.cpp
//
// Library for KSC encodings
// 
// Code adapted from the GNU ICONV Library.
// 

/* Copyright (C) 1999-2000 Free Software Foundation, Inc.

   The GNU ICONV Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   The GNU ICONV Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with the GNU ICONV Library; see the file COPYING.LIB.  If not,
   write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#include "ufcodec.h"
#include "ufcksc.h"

#include <e32std.h>
#include <e32def.h>
#include <e32uid.h>

// ------------------------------------------------------------

extern "C" const unsigned short * const Cjk3380[];

// --------------------------------------------------------------------

typedef struct {
  unsigned short indx; /* index into big table */
  unsigned short used; /* bitmask of used entries */
} Summary16;

#pragma warning ( disable : 4244 )

#include "ksc5601.h"
#include "cp949.h"
#include "johab_hangul.h"
#include "johab.h"

// --------------------------------------------------------------------

class Uni {
public:
  enum { KIllegal = 0xffff, KIncomplete = 0xf800, KHeadless = 0xf801 };
  static TText16 WChar(const TDesC8 &aDes, TInt &aPos);
  static void AppendUTF8(TDes8 &aUtf, TText16 aWChar);
};

GLDEF_C const TUint8 bytesFromUTF8[256] = {
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1, 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
  2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2, 3,3,3,3,3,3,3,3,4,4,4,4,5,5,5,5};

LOCAL_C const TText16 firstByteMark[7] = {
  0x00, 0xC0, 0xE0, 0xF0, 0xF8, 0xFC, 0 };

// convert UTF8 to Unicode
// special cases:  end of string in UTF8 sequence --> KIncomplete
//                 UTF8 sequence stops prematurely --> KIllegal
//                 UTF8 sequence starts with trailer --> KHeadless

TText16 Uni::WChar(const TDesC8 &aDes, TInt &aPos)
{
  TUint wch = aDes[aPos++];
  if ((wch & 0xc0) == 0x80) {
    // KHeadless
    while (aPos < aDes.Length() && (aDes[aPos] & 0xc0) == 0x80)
      aPos++;
    return KHeadless;
  }
  TInt extraBytes = bytesFromUTF8[wch & 0xff];
  wch -= firstByteMark[extraBytes];
  while (extraBytes--) {
    if (aPos >= aDes.Length())
      return TText16(KIncomplete);
    if ((aDes[aPos] & 0xc0) != 0x80)
      // return TText16(aDes[aPos - 1]);
      return TText16(KIllegal);
    wch <<= 6;
    wch |= aDes[aPos++] & 0x3f; 
  }
  if (wch == 0x7fe) return 0xc6d4; // Hangul weol 
  if (wch == 0x7ff) return 0x6708; // moon ideograph
  return TText16(wch);
}

void Uni::AppendUTF8(TDes8 &aUtf, TText16 wch)
{
  if (wch < 0x80) {
    aUtf.Append(TChar(wch));
  } else if (wch < 0x800) {
    aUtf.Append(TChar(((wch & 0x7c0) >> 6) | 0xc0));
    aUtf.Append(TChar((wch & 0x03f) | 0x80));
  } else {
    // We never need to write UCS larger than 0x10000,
    // so now it's three bytes 
    aUtf.Append(TChar(((wch & 0x0f000) >> 12) | 0xe0));
    aUtf.Append(TChar(((wch & 0xfc0) >> 6) | 0x80));
    aUtf.Append(TChar((wch & 0x03f) | 0x80));
  }
}

// ------------------------------------------------------------

class CKscCodec : public CUniCodec {
public:
  virtual HBufC8 *EncodeTextL(const TDesC8 &aText, TBool aSkipUnknown);
  virtual HBufC8 *DecodeTextL(const TDesC8 &aText);
  virtual TInt ApiLevel() const { return 17; }
  TBool PutL(MUniCodecUnicodeSink &sink, TUniChar wch);
};

// --------------------------------------------------------------------

TBool CKscCodec::PutL(MUniCodecUnicodeSink &sink, TUniChar wch)
{
  TBool terminate = EFalse; // satisfy compiler
  if (0x3380 <= wch && wch < 0x33de) {
    wch -= 0x3380;
    const TText16 *ptr = Cjk3380[wch];
    while (*ptr)
      terminate = sink.PutL(*ptr++);
  } else {
    terminate = sink.PutL(wch);
  }
  return terminate;
}

// ------------------------------------------------------------

class TUnicodeSource : public MUniCodecUnicodeSource {
public:
  TUnicodeSource(const TPtrC &aPtr) : iPtr(aPtr), iIndex(0) { }
  TInt GetL();
private:
  const TPtrC &iPtr;
  TInt iIndex;
};

TInt TUnicodeSource::GetL()
{
  if (iIndex < iPtr.Length())
     return Uni::WChar(iPtr, iIndex);
  else
     return -1;
}

class TNativeSink : public MUniCodecNativeSink {
public:
  TNativeSink(TPtr &aPtr, TBool aSkip) : iPtr(aPtr), iSkip(aSkip) { } 
  TBool PutL(TInt byte);
  TBool LossyL(TUniChar wc);
private:
  TPtr &iPtr;
  TBool iSkip;
};

TBool TNativeSink::PutL(TInt byte)
{
  iPtr.Append(TChar(byte));
  return EFalse;
}

TBool TNativeSink::LossyL(TUniChar wc)
{
  if (!iSkip) {
    iPtr.AppendFormat(_L("\\u%04x"), wc);
  }
  return EFalse;
}

// --------------------------------------------------------------------

HBufC8 *CKscCodec::EncodeTextL(const TDesC8 &aText, TBool aSkipUnknown)
{
  HBufC8 *hbuf = HBufC8::NewLC(aText.Length() * Max(6, MaximumSizeInBytes()));
  TUnicodeSource source(aText);
  TPtr8 sinkbuf(hbuf->Des());
  TNativeSink sink(sinkbuf, aSkipUnknown);
  EncodingContextReset();
  EncodeL(&source, sink);
  EncodeL(0, sink);
  CleanupStack::Pop(); // hbuf;
  return hbuf;
}

// --------------------------------------------------------------------

class TNativeSource : public MUniCodecNativeSource {
public:
  TNativeSource(const TPtrC &aPtr) : iPtr(aPtr), iIndex(0) { }
  virtual TInt GetL(TBool first);
private:
  const TPtrC &iPtr;
  TInt iIndex;
};

TInt TNativeSource::GetL(TBool /* first */)
{
  if (iIndex < iPtr.Length())
    return iPtr[iIndex++];
  else
    return -1;
}

class TUtfSink : public MUniCodecUnicodeSink {
public:
  TUtfSink(TPtr &aPtr) : iPtr(aPtr) { } 
  TBool PutL(TUniChar wc);
  TBool IllegalL(TUint32 bytes);
private:
  TPtr &iPtr;
};

TBool TUtfSink::PutL(TUniChar wch) 
{
  Uni::AppendUTF8(iPtr, wch);
  return EFalse;
}

// ------------------------------------------------------------


TBool TUtfSink::IllegalL(TUint32 /* bytes */)
{
  Uni::AppendUTF8(iPtr, 0xfffd);
  return EFalse;
}

// --------------------------------------------------------------------

HBufC8 *CKscCodec::DecodeTextL(const TDesC8 &aText)
{
  HBufC8 *hbuf = HBufC8::NewLC(aText.Length() * 6);
  TNativeSource source(aText);
  TPtr8 sinkbuf(hbuf->Des());
  TUtfSink sink(sinkbuf);
  DecodingContextReset();
  DecodeL(&source, sink);
  DecodeL(0, sink);
  CleanupStack::Pop(); // hbuf;
  return hbuf;
}

// --------------------------------------------------------------------

class CEucKrEncoder : public CKscCodec {
public:
  virtual TText16 DecodeWChar(TText16 wc) { return FromCp949(wc); }
  virtual TText16 EncodeWChar(TText16 wc) { return ToKsc5601(wc); }
  virtual TUid Uid() const { return TUid::Uid(0x1000185d); }
  virtual TTime Version() const
  { return TTime(TDateTime(2000, ENovember, 1, 0, 0, 0, 0)); }
  virtual TInt MaximumSizeInBytes() const { return 2; }
  virtual TInt MaximumSizeOfUnicodeRepresentation() const { return 6; }
  virtual void EncodingContextReset() { }
  virtual TBool EncodingContextNonTrivial() const { return EFalse; }
  virtual TInt EncodingContextData() const { return 0; }
  virtual void DecodingContextReset() { iChar = 0; } 
  virtual TBool DecodingContextNonTrivial() const { return (iChar != 0); }
  virtual TInt DecodingContextData() const 
    { return (iChar != 0) ? 1 : 0; }
  virtual void EncodeL(MUniCodecUnicodeSource *source,
		       MUniCodecNativeSink &sink);
  virtual TInt DecodeL(MUniCodecNativeSource *source,
                       MUniCodecUnicodeSink &sink);
private:
  TInt iChar;
};

// ------------------------------------------------------------

void CEucKrEncoder::EncodeL(MUniCodecUnicodeSource *source, 
			    MUniCodecNativeSink &sink)
{
  if (!source)
    // no context to flush
    return;

  for (;;) {
    TInt wch = source->GetL();
    if (wch < 0) 
      return;
    TText16 aEnc = ToKsc5601(wch);
    if (aEnc) {
      if (aEnc >= 0x100)
	sink.PutL(aEnc >> 8);
      if (sink.PutL(aEnc & 0xff))
	return;
    } else {
      if (sink.LossyL(wch))
	return;
    }
  }
}

TInt CEucKrEncoder::DecodeL(MUniCodecNativeSource *source, 
			    MUniCodecUnicodeSink &sink)
{
  if (!source) {
    if (iChar != 0)
      sink.IllegalL(iChar);
    return 0;
  }
  
  for (;;) {
    if (!iChar) {
      TInt ch = source->GetL(ETrue);
      if (ch < 0)
	return 0;
      if (ch < 0x80) {
	if (PutL(sink, FromCp949(ch)))
	  return 0;
      } else {
	iChar = ch << 8;
      }
    } else {
      // iChar != 0
      TInt ch = source->GetL(EFalse);
      if (ch < 0)
	return 1;
      ch |= iChar;
      iChar = 0;
      if (PutL(sink, FromCp949(ch)))
	return 0;
    }
  }
}

EXPORT_C CUniCodec* NewEucKrEncoder()
{
  return new CEucKrEncoder;
}

// --------------------------------------------------------------------

class CCp949Encoder : public CKscCodec {
public:
  virtual TText16 DecodeWChar(TText16 wc) { return FromCp949(wc); }
  virtual TText16 EncodeWChar(TText16 wc) { return ToCp949(wc); }
  virtual TUid Uid() const { return TUid::Uid(0x1000186b); }
  virtual TTime Version() const
  { return TTime(TDateTime(2000, ENovember, 1, 0, 0, 0, 0)); }
  virtual TInt MaximumSizeInBytes() const { return 2; }
  virtual TInt MaximumSizeOfUnicodeRepresentation() const { return 6; }
  virtual void EncodingContextReset() { }
  virtual TBool EncodingContextNonTrivial() const { return EFalse; }
  virtual TInt EncodingContextData() const { return 0; }
  virtual void DecodingContextReset() { iChar = 0; } 
  virtual TBool DecodingContextNonTrivial() const { return (iChar != 0); }
  virtual TInt DecodingContextData() const 
    { return (iChar != 0) ? 1 : 0; }
  virtual void EncodeL(MUniCodecUnicodeSource *source,
		       MUniCodecNativeSink &sink);
  virtual TInt DecodeL(MUniCodecNativeSource *source,
                       MUniCodecUnicodeSink &sink);
private:
  TInt iChar;
};

// ------------------------------------------------------------

void CCp949Encoder::EncodeL(MUniCodecUnicodeSource *source, 
			    MUniCodecNativeSink &sink)
{
  if (!source)
    // no context to flush
    return;

  for (;;) {
    TInt wch = source->GetL();
    if (wch < 0) 
      return;
    TText16 aEnc = ToCp949(wch);
    if (aEnc) {
      if (aEnc >= 0x100)
	sink.PutL(aEnc >> 8);
      if (sink.PutL(aEnc & 0xff))
	return;
    } else {
      if (sink.LossyL(wch))
	return;
    }
  }
}

TInt CCp949Encoder::DecodeL(MUniCodecNativeSource *source, 
			    MUniCodecUnicodeSink &sink)
{
  if (!source) {
    if (iChar != 0)
      sink.IllegalL(iChar);
    return 0;
  }
  
  for (;;) {
    if (!iChar) {
      TInt ch = source->GetL(ETrue);
      if (ch < 0)
	return 0;
      if (ch < 0x80) {
	if (PutL(sink, FromCp949(ch)))
	  return 0;
      } else {
	iChar = ch << 8;
      }
    } else {
      // iChar != 0
      TInt ch = source->GetL(EFalse);
      if (ch < 0)
	return 1;
      ch |= iChar;
      iChar = 0;
      if (PutL(sink, FromCp949(ch)))
	return 0;
    }
  }
}

EXPORT_C CUniCodec* NewCp949Encoder()
{
  return new CCp949Encoder;
}

// --------------------------------------------------------------------

class CJohabEncoder : public CKscCodec {
public:
  virtual TText16 DecodeWChar(TText16 wc) { return FromJohab(wc); }
  virtual TText16 EncodeWChar(TText16 wc) { return ToJohab(wc); }
  virtual TUid Uid() const { return TUid::Uid(0x1000186b); }
  virtual TTime Version() const
  { return TTime(TDateTime(2000, ENovember, 1, 0, 0, 0, 0)); }
  virtual TInt MaximumSizeInBytes() const { return 2; }
  virtual TInt MaximumSizeOfUnicodeRepresentation() const { return 6; }
  virtual void EncodingContextReset() { }
  virtual TBool EncodingContextNonTrivial() const { return EFalse; }
  virtual TInt EncodingContextData() const { return 0; }
  virtual void DecodingContextReset() { iChar = 0; } 
  virtual TBool DecodingContextNonTrivial() const { return (iChar != 0); }
  virtual TInt DecodingContextData() const 
    { return (iChar != 0) ? 1 : 0; }
  virtual void EncodeL(MUniCodecUnicodeSource *source,
		       MUniCodecNativeSink &sink);
  virtual TInt DecodeL(MUniCodecNativeSource *source,
                       MUniCodecUnicodeSink &sink);
private:
  TInt iChar;
};

// ------------------------------------------------------------

void CJohabEncoder::EncodeL(MUniCodecUnicodeSource *source, 
			    MUniCodecNativeSink &sink)
{
  if (!source)
    // no context to flush
    return;

  for (;;) {
    TInt wch = source->GetL();
    if (wch < 0) 
      return;
    TText16 aEnc = ToJohab(wch);
    if (aEnc) {
      if (aEnc >= 0x100)
	sink.PutL(aEnc >> 8);
      if (sink.PutL(aEnc & 0xff))
	return;
    } else {
      if (sink.LossyL(wch))
	return;
    }
  }
}

TInt CJohabEncoder::DecodeL(MUniCodecNativeSource *source, 
			    MUniCodecUnicodeSink &sink)
{
  if (!source) {
    if (iChar != 0)
      sink.IllegalL(iChar);
    return 0;
  }
  
  for (;;) {
    if (!iChar) {
      TInt ch = source->GetL(ETrue);
      if (ch < 0)
	return 0;
      if (ch < 0x80) {
	if (PutL(sink, FromJohab(ch)))
	  return 0;
      } else {
	iChar = ch << 8;
      }
    } else {
      // iChar != 0
      TInt ch = source->GetL(EFalse);
      if (ch < 0)
	return 1;
      ch |= iChar;
      iChar = 0;
      if (PutL(sink, FromJohab(ch)))
	return 0;
    }
  }
}

EXPORT_C CUniCodec* NewJohabEncoder()
{
  return new CJohabEncoder;
}

// ------------------------------------------------------------

class CIsoKrEncoder : public CKscCodec {
public:
  virtual TText16 DecodeWChar(TText16 wc);
  virtual TText16 EncodeWChar(TText16 wc);
  virtual TUid Uid() const { return TUid::Uid(0x1000186c); }
  virtual TTime Version() const
  { return TTime(TDateTime(2000, ENovember, 4, 0, 0, 0, 0)); }
  virtual TInt MaximumSizeInBytes() const { return 7; }
  virtual TInt MaximumSizeOfUnicodeRepresentation() const { return 6; }
  virtual void EncodingContextReset() 
  { iHeaderWritten = iEncodingContextHanja = EFalse; }
  virtual TBool EncodingContextNonTrivial() const
  { return iHeaderWritten || iEncodingContextHanja; }
  virtual TInt EncodingContextData() const { return 0; }
  virtual void DecodingContextReset()
  { iChar = 0; iPending = 0; iDecodingContextHanja = EFalse; } 
  virtual TBool DecodingContextNonTrivial() const 
  { return (iPending != 0) || (iChar != 0) || iDecodingContextHanja; }
  virtual TInt DecodingContextData() const 
  { return iPending + (iChar != 0 ? 1 : 0); }
  virtual void EncodeL(MUniCodecUnicodeSource *source,
		       MUniCodecNativeSink &sink);
  virtual TInt DecodeL(MUniCodecNativeSource *source,
                       MUniCodecUnicodeSink &sink);
private:
  void FlushPartialSequence(MUniCodecUnicodeSink &sink);
private:
  TBool iHeaderWritten;
  TBool iEncodingContextHanja;
  TBool iDecodingContextHanja;
  TInt iPending;
  TInt iChar;
};

// ------------------------------------------------------------

LOCAL_C TText16 FilterEucKr(TText16 wc)
{
  if (0xa1a1 <= wc && wc <= 0xfefe) {
    TText8 c2 = wc & 0xff;
    if (0xa1 <= c2 && c2 <= 0xfe) {
      return wc - 0x8080;
    }
  }
  return 0;
}

TText16 CIsoKrEncoder::EncodeWChar(TText16 wc)
{
  if (wc < 0x80) return wc;
  return FilterEucKr(ToKsc5601(wc));
}

TText16 CIsoKrEncoder::DecodeWChar(TText16 wc)
{
  if (wc < 0x80) return wc;
  wc += 0x8080;
  TText16 check = FilterEucKr(wc);
  if (check)
    return FromKsc5601(wc);
  return 0;
}

void CIsoKrEncoder::EncodeL(MUniCodecUnicodeSource *source, 
			    MUniCodecNativeSink &sink)
{
  if (!source) {
    if (iEncodingContextHanja) {
      // single byte mode
      sink.PutL(0x0f);
    }
    return;
  }

  if (!iHeaderWritten) {
    sink.PutL(0x1b); // ESC
    sink.PutL('$');
    sink.PutL(')');
    sink.PutL('C');
    iHeaderWritten = ETrue;
  }

  for (;;) {
    TInt wch = source->GetL();
    if (wch < 0) 
      return;
    if (wch < 0x80) {
      if (iEncodingContextHanja) {
	iEncodingContextHanja = EFalse;
	// single byte mode
	sink.PutL(0x0f);
      }
      if (sink.PutL(wch))
	return;
    } else {
      TText16 aEnc = FilterEucKr(ToKsc5601(wch));
      if (aEnc) {
	if (!iEncodingContextHanja) {
	  iEncodingContextHanja = ETrue;
	  // double byte mode
	  sink.PutL(0x0e);
	}
	sink.PutL(aEnc >> 8);
	if (sink.PutL(aEnc & 0xff))
	  return;
      } else {
	if (iEncodingContextHanja) {
	  iEncodingContextHanja = EFalse;
	  // single byte mode
	  sink.PutL(0x0f);
	}
	if (sink.LossyL(wch))
	  return;
      }
    }
  }
}

void CIsoKrEncoder::FlushPartialSequence(MUniCodecUnicodeSink &sink)
{
  // flush out partial control sequence
  if (iPending--)
    sink.PutL(0x1b);
  if (iPending--)
    sink.PutL('$');
  if (iPending--)
    sink.PutL(')');
}

TInt CIsoKrEncoder::DecodeL(MUniCodecNativeSource *source, 
			    MUniCodecUnicodeSink &sink)
{
  if (!source) {
    if (iPending)
      FlushPartialSequence(sink);
    if (iChar)
      sink.IllegalL(iChar);
    return 0;
  }
  
  for (;;) {
    if (!iPending && !iChar) {
      TInt ch = source->GetL(ETrue);
      if (ch < 0)
	return 0;
      if (ch == 0x0e) {
	// double byte mode
	iDecodingContextHanja = ETrue;
      } else if (ch == 0x0f) {
	// single byte mode
	iDecodingContextHanja = EFalse;
      } else if (ch == 0x1b) {
	// wait for ESC $ ) C
	iPending = 1;
      } else if (ch <= 0x20 || !iDecodingContextHanja) {
	if (sink.PutL(ch))
	  return 0;
      } else if (0x21 <= ch && ch <= 0x7e) {
	// interpret two bytes
	iChar = ch << 8;
      } else {
	if (sink.IllegalL(ch))
	  return 0;
	
      }
    } else if (iPending) {
      // in ESC $ ) C sequence
      TInt ch = source->GetL(EFalse);
      if (ch < 0)
	return 0; // not in middle of character!
      if ((iPending == 1 && ch == '$') ||
	  (iPending == 2 && ch == ')') ||
	  (iPending == 3 && ch == 'C')) {
	iPending++;
      } else {
	// different sequence
	FlushPartialSequence(sink);
	if (sink.PutL(ch))
	  return 0;
      }
      if (iPending == 4)
	// sequence recognized -> drop it
	iPending = 0;
    } else {
      // waiting for second byte
      TInt ch = source->GetL(EFalse);
      if (ch < 0)
	return 1;
      ch |= iChar;
      iChar = 0;
      if (PutL(sink, FromKsc5601(ch + 0x8080)))
	return 0;
    }
  }
}

EXPORT_C CUniCodec* NewIsoKrEncoder()
{
  return new CIsoKrEncoder;
}

// ------------------------------------------------------------

// The E32Dll() entry point function

GLDEF_C TInt E32Dll(TDllReason /*aReason*/)
{
  return(KErrNone);
}
